#!/usr/bin/perl -w
#
# 65816.pl
#
# A 65816 simulator.  The main purpose of this is to be used as
# a wrapper for running Slang on a PC/Linux.
#
# This software is freely distributable.
#
# Stephen L. Judd
# 2/2/05
#
# 3/19/05 This is totally slow for running stuff, so a routine has
#	  been added to make a C header file for a quick hacked up C version.
#
#	  Known bugs: LoadROM loads to RAM, not ROM.  ReadByte needs
#	  have IsReg check (like FetchMem).
#

#use strict;

# RAM/ROM/Regs

#our @rambank = ("\0") x 0x10000;
our @rom = (0) x 0x10000;
our @rambank = (0) x 0x10000;
our %ram;			# More efficient as a hash
$ram{0} = [@rambank];

# CPU registers

our $reg_a = 0;
our $reg_x = 0;
our $reg_y = 0;
our $reg_pc = 0;
our $reg_dbr = 0;
our $reg_pbr = 0;
our $reg_sp = 0x01ff;
our $reg_flags = 0;
our $reg_e = 1;
our $reg_d = 0;			# Direct page

# Opcode matrix

my @oplist = qw/
  BRK s  ORA (d,x)  COP s  ORA d,s  TSB d  ORA d  ASL d  ORA [d]
	PHP s  ORA #  ASL A  PHD s  TSB a  ORA a  ASL a  ORA al
  BPL r  ORA (d),y  ORA [d]  ORA (d,s),y  TRB d  ORA d,x  ASL d,x  ORA [d],y
	CLC i  ORA a,y  INC A  TCS i  TRB a  ORA a,x  ASL a,x  ORA al,x
  JSR a  AND (d,x)  JSL al  AND d,s  BIT d  AND d  ROL d  AND [d]
	PLP s  AND #  ROL A  PLD s  BIT a  AND a  ROL a  AND al
  BMI r  AND (d),y  AND (d)  AND (d,s),y  BIT d,x  AND d,x  ROL d,x  AND [d],y
	SEC i  AND a,y  DEC A  TSC i  BIT a,x  AND a,x  ROL a,x  AND al,x
  RTI s  EOR (d,x)  WDM i  EOR d,s  MVP xyc  EOR d  LSR d  EOR [d]
	PHA s  EOR #  LSR A  PHK s  JMP a  EOR a  LSR a  EOR al
  BVC r  EOR (d),y  EOR (d)  EOR (d,s),y  MVN xyc  EOR d,x  LSR d,x  EOR [d],y
	CLI i  EOR a,y  PHY s  TCD i  JMP al  EOR a,x  LSR a,x  EOR al,x
  RTS s  ADC (d,x)  PER rl  ADC d,s  STZ d  ADC d  ROR d  ADC [d]
	PLA s  ADC #  ROR A  RTL s  JMP (a)  ADC a  ROR a  ADC al
  BVS r  ADC (d),y  ADC (d)  ADC (d,s),y  STZ d,x  ADC d,x  ROR d,x  ADC [d],y
	SEI i  ADC a,y  PLY s  TDC i  JMP (a,x)  ADC a,x  ROR a,x  ADC al,x
  BRA r  STA (d,x)  BRL rl  STA d,s  STY d  STA d  STX d  STA [d]
	DEY i  BIT #  TXA i  PHB s  STY a  STA a  STX a  STA al
  BCC r  STA (d),y  STA (d)  STA (d,s),y  STY d,x  STA d,x  STX d,y  STA [d],y
	TYA i  STA a,y  TXS i  TXY i  STZ a  STA a,x  STZ a,x  STA al,x
  LDY #  LDA (d,x)  LDX #  LDA d,s  LDY d  LDA d  LDX d  LDA [d]
	TAY i  LDA #  TAX i  PLB s  LDY a  LDA a  LDX a  LDA al
  BCS r  LDA (d),y  LDA (d)  LDA (d,s),y  LDY d,x  LDA d,x  LDX d,y  LDA [d],y
	CLV i  LDA a,y  TSX i  TYX i  LDY a,x  LDA a,x  LDX a,y  LDA al,x
  CPY #  CMP (d,x)  REP #  CMP d,s  CPY d  CMP d  DEC d  CMP [d]
	INY i  CMP #  DEX i  WAI i  CPY a  CMP a  DEC a  CMP al
  BNE r  CMP (d),y  CMP (d)  CMP (d,s),y  PEI (d)  CMP d,x  DEC d,x  CMP [d],y
	CLD i  CMP a,y  PHX s  STP i  JML (a)  CMP a,x  DEC a,x  CMP al,x
  CPX #  SBC (d,x)  SEP #  SBC d,s  CPX d  SBC d  INC d  SBC [d]
	INX i  SBC #  NOP i  XBA i  CPX a  SBC a  INC a  SBC al
  BEQ r  SBC (d),y  SBC (d)  SBC (d,s),y  PEA s  SBC d,x  INC d,x  SBC [d],y
	SED i  SBC a,y  PLX s  XCE i  JSR (a,x)  SBC a,x  INC a,x  SBC al,x/;


my %modelen = (
	"#" => 1,	# Immediate
	A => 0,		# Accumulator
	i => 0,		# implied
	s => 0,		# stack
	r => 1,		# relative
	rl => 2,	# relative long
	d => 1,		# direct page
	"d,x" => 1,
	"d,y" => 1,
	"(d)" => 1,
	"(d,x)" => 1,
	"(d),y" => 1,
	"[d]" => 1,
	"[d],y" => 1,
	a => 2,		# absolute
	"a,x" => 2,
	"a,y" => 2,
	al => 3,	# absolute long
	"al,x" => 3,
	"d,s" => 1,	# stack relative
	"(d,s),y" => 1,
	"(a)" => 2,	# absolute indirect
	"(a,x)" => 2,
	xyc => 2	# block move
  );

#
# Available memory configurations are:
#
#  lower 3 bits
# 	01	$A000	$D000	$E000
#       --	-----	-----	-----
#	x00	RAM	RAM	RAM
#	001	RAM	ChrROM	RAM
#	101	RAM	I/O	RAM
#	010	RAM	ChrROM	Kernal
#	110	RAM	I/O	Kernal
#	011	BASIC	ChrROM	Kernal
#	111	BASIC	I/O	Kernal
#

my @a000 = qw/ram ram ram rom ram ram ram rom/;
my @d000 = qw/ram rom rom rom ram io io io/;
my @e000 = qw/ram ram rom rom ram ram rom rom/;


my $opcode;	# Opcode byte read in
my $mnem;
my $mode;
my $oppc;			# PC of opcode being read in
my @address = (0, 0, 0);	# Physical bytes read in
my $effective_address=0;
my @bytesread;			# All bytes read by ParseOp

my $emul=1;			# Emulation mode (1=emulation mode)
my $xflag=1;			# X processor flag
my $mflag=1;			# M processor flag
my $cflag=0;
my $alo=0;			# low byte, .A reg
my $ahi=0;
my $xlo=0;
my $xhi=0;
my $ylo=0;
my $yhi=0;

my $debug_iter=0;


#&MakeCFiles;
#exit;

&Init65816;
#my ($loadaddr, $endaddr) = &LoadRAM("SPRITEDEMO.O", 0);
#print "Load addr=$loadaddr, End addr=$endaddr\n";
#&IterateCode($loadaddr, 80);

print "Loading ROM/RAM...\n";
&LoadROM("bin/rom-a000.prg", 0xa000);
&LoadROM("bin/rom-e000.prg", 0xe000);
&LoadROM("bin/CHARROM-D000.PRG", 0xd000);
&LoadRAM("bin/ZP0000-03FF.PRG", 0);
&LoadRAM("bin/SLANGB1.10.O", 0);
&LoadRAM("bin/slangdef.o", 0);

my ($start, $end) = &LoadRAM("SPRITEDEMO.E.S", 0x040000);
$effective_address = $end;
$mode = "a";		# For StoreMem
&StoreMem(0, 1);	# Store a 0 at end of text
print "Source code at \$", &ToHex24($start), "-\$", &ToHex24($end), "\n";
print "Compiling...\n";
&SetAXY($start);
&CallAsSub(0x0810);	# Init TEXTP
&SetAXY($end+1);
&CallAsSub(0x0813);	# Init TEXTEND
&CallAsSub(0x0816);	# Assemble
&CallAsSub(0x0819);	# GetORG
my $org = &GetAXY;
&CallAsSub(0x081c);	# GetCODELEN; assembled code at $020000-$02CODELEN
my $codelen = &GetAXY;

print "\nORG=\$", &ToHex16($org), ", CODEEND=\$", &ToHex16($codelen), "\n";
&SaveObj("slangout.o", $org, $codelen);

#&IterateCode(0x080d, 80);
  
exit;

sub DisplayMem {
  my ($start, $end) = shift;
  my $i = $start;
  while ($i <= $end) {
    my $str = "";
    print &ToHex16($i), ">";
    for my $count (0..15) {
      my $byte = $ram{0}[$i];
      if (($byte & 0x7f) >= 32) {
        $str .= chr($byte & 0x7f);
      } else {
        $str .= ".";
      }
      print " ", &ToHex($byte);
      $i++;
    }
    print " $str\n";
  }
}

#
# Test code
#

sub SimpleTest {
  &Init65816;
  my $i=0x1000;

  #foreach (0x20, 04, 0x10, 0x60, 0xa9, 01, 0x8d, 0x20, 0xd0, 0x60) {

  #foreach (0x18, 0xa9, 0xff, 0x69, 0x02, 0xb0, 0xfc, 0x0a, 0x00) {
  foreach (0xa9, 0x00, 0x85, 0x02, 0xa9, 0x10, 0x85, 0x03, 0xa0, 0x00,
	   0xb1, 0x02, 0xd9, 0x00, 0x10, 0xc8, 0xc0, 0x04, 0xd0, 0xf6, 0x60) {
  #foreach (0xad, 0x00, 0xb0, 0xad, 0x20, 0xd0, 0xad, 0xd2, 0xff) {
    $ram{0}[$i] = $_;
    $i += 1;
  }

  #&IterateCode(0x1000, 50);
  &CallAsSub(0x1000);
}

sub IterateCode {
  my $addr = shift;
  my $iter = shift;
  my $i;

  &SetPC($addr);
  for $i (1..$iter) {
    &ExOp;
#    &PrintDisassembly;
  }
}

sub PrintDisassembly {
    my @dis = &DisasOp;
    my $txt = pop @dis;
    print &ToHex16($oppc),"> ";
    for my $j (0..3) {
      if (exists $dis[$j]) {
        print &ToHex($dis[$j]), " ";
      } else {
        print "   ";
      }
    }
    print sprintf("%-11s", $txt);
    print " .A=", &ToHex16($reg_a);
    print " .X=", &ToHex16($reg_x);
    print " .Y=", &ToHex16($reg_y);
    print " SR=", &ToHex($reg_flags);
    print " SP=", &ToHex16($reg_sp), ">";
    for my $j (0..3) {
      print &ToHex($ram{0}[$reg_sp+$j]), " ";
    }
    #print "  d020=$ram{0}[0xd020]\n";
    print "\n";
}

#
# MakeCFiles makes a little header file that can be used by
# a C program.
#
sub MakeCFiles {
  my %opcodes;
  my %defmode = (
	"#" => _IMMED_,		# Immediate
	A => _ACCUM_,		# Accumulator
	i => _IMPLIED_,		# implied
	s => _STACK_,		# stack
	r => _REL_,		# relative
	rl => _LONGREL_,	# relative long
	d => _DP_,		# direct page
	"d,x" => _DPX_,
	"d,y" => _DPY_,
	"(d)" => _IDP_,
	"(d,x)" => _IDPX_, 
	"(d),y" => _IDPY_,
	"[d]" => _LIDP_,
	"[d],y" => _LIDPY_,
	a => _ABS_,		# absolute
	"a,x" => _ABSX_,
	"a,y" => _ABSY_,
	al => _LONGABS_,	# absolute long
	"al,x" => _LONGABSX_,
	"d,s" => _STACKREL_,	# stack relative
	"(d,s),y" => _STACKRELY_,
	"(a)" => _IABS_,	# absolute indirect
	"(a,x)" => _IABSX_,
	xyc => _XYC_	# block move
  );

  for my $k (0..255) {
    $opcodes{$oplist[2*$k]} = 1;
  }

  open FH, ">opdefs.h";
  open GH, ">opglobs.h";
  
  print FH "/*\n * Opcode macros.";
  print FH "  Do not modify -- create with 65816.pl.\n */\n\n";
  print GH "/*\n * Constant, global opcode arrays.";
  print GH "  Do not modify -- create with 65816.pl.\n */\n\n";
  print GH '#include "opdefs.h"' . "\n\n";


  my $k = 0;
  foreach (sort keys %opcodes) {
    print FH "#define $_ $k\n";
    $k++;
  }
  print FH "\n\n";

  $k=0;
  foreach (sort keys %defmode) {
    print FH "#define $defmode{$_} $k\n";
    $k++;
  }
  print FH "\n\n";

  print GH "unsigned char oplist[512] = {\n";
  for $k (0..255) {
    my $op = $oplist[2*$k];
    my $mode = $oplist[2*$k+1];
    print GH "  $op, $defmode{$mode},\t/* \$", &ToHex($k), " */\n";
  }
  print GH "};\n\n";

  print GH "unsigned char modelen[24] = {\n";
  $k = 0;
  foreach (sort keys %defmode) {	# This keeps stuff in same order
 					# as #defines above.
    print GH "  $modelen{$_}, \t/*\t$_\t\t$defmode{$_} */\n";
    $k++;
  }
  print GH "};\n\n";

  print GH "char opcodes[92][4] = {\n";
  foreach (sort keys %opcodes) {
    print GH '"', "$_", '",', "\n";	# Prints erroneous last comma
  }					# yucky but won't hurt anything...
  print GH "};";

  close FH;
  close GH;
}

#
# CallAsSub is the main code execution routine.  It pretends to call
# the address parameter as a JSR, by setting PC=$FFFE and waiting for
# an RTS to set it back.
#
# This code also intercepts some of the kernal routines, including
# FFD2, OPEN, LOAD, etc.
#
sub CallAsSub {
  my $addr = shift;

  $effective_address = $addr;
  &SetPC(0xffff);
  &JSR;
  while ($reg_pc < 0xffff) {
    $debug_iter++;
    &ExOp;
    #&PrintDisassembly;
    if ($reg_pc == 0xffd2) {	# CHROUT
      my $chr;
      $chr = chr($reg_a & 0x7f);
      if ($reg_a < 32) {	# Could sorta use tr, but...
	$chr = "\n";
      } elsif (($reg_a > 64) && ($reg_a < 92)) {
	$chr = chr($reg_a + 32);
      } elsif (($reg_a > 96) && ($reg_a < 128)) {
	$chr = chr($reg_a - 32);
      } elsif (($reg_a > 192) && ($reg_a < 224)) {
	$chr = chr($reg_a - 128);
      }
      print $chr;
      #if ($chr < 128) {
#	if (ord($chr) == 13) {
#	  print "\n";
#	} elsif (ord($chr) >= 32) {
#	  print $chr;
#	}
      #}
      &RTS;
    }

#    if (($ram{0}[0x2b1d] != 0) || ($debug_iter > 185660)) {&PrintDisassembly};
  }
}


#
# The Sirius interface uses .AXY to set TEXTP and TEXTEND.
#

sub SetAXY {
  my $arg = shift;
  $reg_a = $arg & 0xff;
  $reg_x = (int $arg >> 8) & 0xff;
  $reg_y = (int $arg >> 16) & 0xff;
}

sub GetAXY {
  return $reg_a + 256*($reg_x + 256*$reg_y);
}

#
# Load files to ram/rom
#
# Input:
#   filename
#   load address   x -> load to address x, skip 1st two bytes of file
#		  -x -> load to x, do not skip 1st two bytes
#		   0 -> use file load address (1st two bytes of file)
#   petscii flag (optional) - =1 means strip LF/convert to CR.
#
# Returns:
#   Load start address, load end address+1
#
sub LoadRAM {
  my $fname = shift;
  my $addr = shift;
  my $petscii = shift;

  my $byte;
  my $faddr;
  $mode = "a";		# For StoreMem
  open(INFILE, "<$fname") or die("Can't open $fname: $!\n");
  binmode INFILE;

  if ($addr >= 0) {
    sysread(INFILE, $byte, 1);
    $faddr = ord($byte);
    sysread(INFILE, $byte, 1);
    $faddr = $faddr + 256*ord($byte);
    if ($addr == 0) {
      $addr = $faddr;
    }
  } else {
    $addr = -$addr;
  }

  $effective_address = $addr;
  while (sysread(INFILE, $byte, 1)) {
    $byte = ord($byte);
    if (($petscii) && ($petscii > 0)) {
      if ($byte < 32) {
	$byte = 13;		# Convert all to CR
      }
    }
    &StoreMem($byte, 1);
    $effective_address++;

#    print "$addr> ";
#    print "read byte ", $byte+0, " = ";
#    print &ToHex($byte+0), "\n";

#    $ram{0}[$i] = $byte;
#    $i++;
  }
  close INFILE;
  return $addr, $effective_address;
}

sub LoadROM {
  my $fname = shift;
  my $addr = shift;

  my $byte;
  my $faddr;
  $mode = "a";		# For StoreMem
  open(INFILE, "<$fname") or die("Can't open $fname: !$\n");
  binmode INFILE;

  if ($addr >= 0) {
    sysread(INFILE, $byte, 1);
    $faddr = ord($byte);
    sysread(INFILE, $byte, 1);
    $faddr = $faddr + 256*ord($byte);
    if ($addr == 0) {
      $addr = $faddr;
    }
  }

  $effective_address = $addr;
  while (sysread(INFILE, $byte, 1)) {
    $byte = ord($byte);
    &StoreMem($byte, 1);
    $effective_address++;
  }
  close INFILE;
  return $addr, $effective_address;
}

#
# Save the object code at $020000, $len bytes total
#
sub SaveObj {
  my $fname = shift;
  my $org = shift;
  my $len = shift;

  my $byte;
  my $faddr;
  $mode = "a";		# For FetchMem
  open(OUTFILE, ">$fname") or die("Can't open $fname: $!\n");
  binmode OUTFILE;

  print OUTFILE chr($org & 0xff);
  print OUTFILE chr((int $org/256) & 0xff);
  #syswrite(OUTFILE, ord($org & 0xff), 1);
  #syswrite(OUTFILE, ord(int $org/256), 1);

  for my $count (0..$len-1) {
    $effective_address = 0x020000+$count;
    $byte = &FetchMem(1);
    print OUTFILE chr($byte & 0xff);
  }
  close OUTFILE;
}


#
# Init
#
sub Init65816 {
  $ram{0} = [@rambank];
  $ram{0}[1] = 0x00;	# Force default $01 = all RAM.
}

#
# Execute an opcode: read opcode and address, call routine.
#
sub ExOp {
  $oppc = $reg_pc;
  &ParseOp;
  &{$mnem};
  &FixRegs;
}

sub ParseOp {
  my $i;

  @address = (0, 0, $reg_dbr);
  @bytesread = ();

  $opcode = &ReadByte;
  $mnem = $oplist[2*$opcode];
  $mode = $oplist[2*$opcode + 1];
  push @bytesread, $opcode;
  if (!exists $modelen{$mode}) {die "Unknown addressing mode: $mnem $mode\n";}
  my $plus = &CheckModeLen;
  for $i (1..$modelen{$mode}+$plus) {
    $address[$i-1] = &ReadByte;
    push @bytesread, $address[$i-1];
  }
  &FixAddress();
}

#
# If #immed mode, and in native mode, and register flag is set
# appropriately, then need to read an extra byte.
#

sub CheckModeLen {
  if ($emul == 1) {
    return 0;
  }
  if ($mode ne "#") {
    return 0;
  }

  my $val=0;
  if ($mnem =~ /ADC|AND|BIT|CMP|EOR|LDA|ORA|SBC/) {
    if ($mflag == 0) {
      $val=1;
    }
  } elsif ($mnem =~ /CPX|CPY|LDX|LDY/) {
    if ($xflag == 0) {
      $val=1;
    }
  }
  return $val;
}

#
# Read a byte using PBR and PC
# On exit: PC -> PC+1
#

sub ReadByte {
  &CheckRamPBR;
  $reg_pc = $reg_pc + 1;
  return $ram{$reg_pbr}[$reg_pc-1];
}

#
# Fix up an address based on the addressing mode.  The result is a
# 24-bit address.  Note that this is required to handle passing bank
# boundaries correctly (RAM is stored as a hash, not a fixed+flat array).
#

sub FixAddress {
  $effective_address = $address[0] + 256*($address[1] + 256*$address[2]);
  my $zp = $address[0] + 256*$address[1] + $reg_d; # Bank 0, for zp guys
  if ($mode eq "r") {
    if ($address[0] >= 128) {
	$effective_address -= 256;
    }
    $effective_address = $reg_pc + $effective_address;
  } elsif ($mode eq "rl") {
    if ($address[1] >= 128) {
	$effective_address -= 0x10000;
    }
    $effective_address = $reg_pc + $effective_address;
  } elsif ($mode eq "d") {
    $effective_address = $zp;
  } elsif ($mode eq "d,x") {
    $effective_address = $zp + $reg_x;
  } elsif ($mode eq "d,y") {
    $effective_address = $zp + $reg_y;
  } elsif ($mode eq "(d)") {
    $effective_address = &GetZPPtr16(0);
#print "(d) -> \$", &ToHex24($effective_address), " = ",
#chr(&FetchMemA & 0x7f), " \$", &ToHex(&FetchMemA), " iter=$debug_iter\n";
  } elsif ($mode eq "(d,x)") {
    $effective_address = &GetZPPtr16($reg_x);
  } elsif ($mode eq "(d),y") {
    $effective_address = &GetZPPtr16(0) + $reg_y;
  } elsif ($mode eq "[d]") {
    $effective_address = &GetZPPtr24;
  } elsif ($mode eq "[d],y") {
    $effective_address = &GetZPPtr24 + $reg_y;
  } elsif ($mode eq "a,x") {
    $effective_address += $reg_x;
  } elsif ($mode eq "a,y") {
    $effective_address += $reg_y;
  } elsif ($mode eq "al,x") {
    $effective_address += $reg_x;
  } elsif ($mode eq "d,s") {
    $effective_address = ($address[0] + $reg_sp) & 0xffff; # Potential +1 bug
  } elsif ($mode eq "(d,s),y") {
    my $addr = ($address[0] + $reg_sp) & 0xffff;
    my $addr2 = ($address[0] + $reg_sp + 1) & 0xffff;
    my $lo = $ram{0}[$addr];
    my $hi = $ram{0}[$addr2];
    $effective_address = $lo + 256*($hi + 256*$reg_dbr) + $reg_y;
  } elsif ($mode eq "(a)") {
    $effective_address = &GetIndirect16(0);
  } elsif ($mode eq "(a,x)") {
    $effective_address = &GetIndirect16($reg_x);
  }
    
}

sub GetIndirect16 {
  my $offset = shift;
  my $addr = ($effective_address + $offset) & 0xffff;
  my $addr2 = ($effective_address + $offset + 1) & 0xffff;
  my $lo = $ram{0}[$addr];
  my $hi = $ram{0}[$addr2];
  return $lo + 256*($hi + 256*$reg_dbr);
}

sub GetZPPtr16 {
  my $offset = shift;
  my $addr = ($effective_address + $reg_d + $offset) & 0xffff;
  my $addr2 = ($effective_address + $reg_d + $offset + 1) & 0xffff;
  my $lo = $ram{0}[$addr];
  my $hi = $ram{0}[$addr2];
  return $lo + 256*($hi + 256*$reg_dbr);
}

sub GetZPPtr24 {
  my $addr = ($effective_address + $reg_d) & 0xffff;
  my $addr2 = ($effective_address + $reg_d + 1) & 0xffff;
  my $addr3 = ($effective_address + $reg_d + 2) & 0xffff;
  my $lo = $ram{0}[$addr];
  my $hi = $ram{0}[$addr2];
  my $hi24 = $ram{0}[$addr3];
  return $lo + 256*($hi + 256*$hi24);
}

#
# Fix up the register bitness, set flag shortcuts
#

sub FixRegs {
  $emul = $reg_e;
  $cflag = $reg_flags & 01;
  if ($emul == 1) {
    $mflag=1;
    $xflag=1;
    $reg_sp = 0x0100 + ($reg_sp & 0xff);
  } else {
    $mflag = int ($reg_flags & 0x20) / 0x20;
    $xflag = int ($reg_flags & 0x10) / 0x10;
  }

  if (($mflag | $emul) == 1) {
    $reg_a = $reg_a & 255;
    $alo = $reg_a;
  } else {
    $alo = $reg_a & 255;
    $ahi = int $reg_a/256;
  }

  if (($xflag | $emul) == 1) {	# note: in emul mode, xhi = 0
    $reg_x = $reg_x & 255;
    $xlo = $reg_x;
    $xhi = 0;
    $reg_y = $reg_y & 255;
    $ylo = $reg_y;
    $yhi = 0;
  } else {
    $xlo = $reg_x & 255;
    $xhi = int $reg_x/256;
    $ylo = $reg_y & 255;
    $yhi = int $reg_y/256;
  }
}

#
# Set the program counter + pbr
#

sub SetPC {
  my $val = shift;

  $reg_pbr = int $val/65536;
  $reg_pc = $val & 65535;
}

##################################
#
# Disassembly routines
#
##################################

#
# Disassemble current opcode
# Returns string [op arg] and array of bytes
#
sub DisasOp {
  my $str = $mnem . " ";
  my $arg = "\$";
  my @temp = @bytesread;
  while (exists $temp[1]) {
    $arg .= &ToHex(pop @temp);
  }
  my $blah = $mode;
  if ($mode eq "#") {
    $blah = "#$arg";
  } elsif ($mode =~ /r|rl/) {
    $blah = "\$" . &ToHex16($effective_address);
  } elsif ($blah =~ /.*d/) {
    $blah =~ s/d/$arg/;
  } elsif ($blah =~ /.*al/) {
    $blah =~ s/al/$arg/;
  } elsif ($blah =~ /.*a/) {
    $blah =~ s/a/$arg/;
  } else {
    $blah = "";
  }
  $str .= $blah;

  return @bytesread, $str;
}

#
# Convert a number to a 2-digit hex value
#
sub ToHex {
  my $val = shift;
  my @hex = qw/0 1 2 3 4 5 6 7 8 9 a b c d e f/;
  my $hi = int $val/16;
  my $lo = $val - 16*$hi;
  return $hex[$hi] . $hex[$lo];
}

sub ToHex16 {
  my $val = shift;
  return &ToHex(int $val/256) . &ToHex($val - 256*int($val/256));
}

sub ToHex24 {
  my $val = shift;
  my $hi24 = &ToHex(int $val >> 16);
  my $hi = &ToHex((int $val >> 8) & 0xff);
  my $lo = &ToHex($val & 0xff);
  return $hi24 . $hi . $lo;
}

##################################
#
# Opcodes
#
##################################

sub ADC {
  my $argr = &FetchMemA;
  my $argl = $reg_a;
  $reg_a += $argr + $cflag;
  &SetCarry($reg_a);
  &SetV($argl, $argr);
  &SetNeg($reg_a, $mflag);
  &SetZ($reg_a, $mflag);
}

sub AND {
  $reg_a = $reg_a & &FetchMemA;
  &SetNeg($reg_a, $mflag);
  &SetZ($reg_a, $mflag);
}

sub ASL {
  my $blah = &FetchMemA << 1;
  &SetCarry($blah);
  &SetNeg($blah, $mflag);
  &SetZ($blah, $mflag);
  &StoreMem($blah, $mflag);
}

sub BCC {
  &BranchClear(0x01);
}

sub BCS {
  &BranchSet(0x01);
}

sub BEQ {
  &BranchSet(0x02);
}

sub BIT {
  my $mem = &FetchMemA;
  my $arg = $reg_a & $mem;
  my $hi = $mem & 0xc0;		# M7 and M6
  if ($mflag == 0) {
    $hi = $mem & 0xc000;
  }
 
  &SetZ($arg, $mflag);
  if ($mode ne "#") {		# M7->N, M6->V
    $reg_flags = (($reg_flags & 0x3f) | $hi);
  }
}

sub BMI {
  &BranchSet(0x80);
}

sub BNE {
  &BranchClear(0x02);
}

sub BPL {
  &BranchClear(0x80);
}

sub BRA {
  $reg_pc = $effective_address;
}

sub BRK {
  &NotImplemented;
}

sub BRL {
  $reg_pc = $effective_address;
}

sub BVC {
  &BranchClear(0x40);
}

sub BVS {
  &BranchSet(0x40);
}

sub CLC {
  &ClearFlag(0x01);
}

sub CLD {
  &ClearFlag(0x08);
}

sub CLI {
  &ClearFlag(0x04);
}

sub CLV {
  &ClearFlag(0x40);
}

sub CMP {
  &CompareMem($reg_a, $mflag);
}

sub COP {
  &NotImplemented;
}

sub CPX {
  &CompareMem($reg_x, $xflag);
}

sub CPY {
  &CompareMem($reg_y, $xflag);
}

sub DEC {
  my $arg = &Neg16(&FetchMemA - 1);
  &StoreMem($arg, $mflag);
  &SetNeg($arg, $mflag);
  &SetZ($arg, $mflag);
}

sub DEX {
  $reg_x = &Neg16($reg_x - 1);
  &SetNeg($reg_x, $xflag);
  &SetZ($reg_x, $xflag);
}

sub DEY {
  $reg_y = &Neg16($reg_y - 1);
  &SetNeg($reg_y, $xflag);
  &SetZ($reg_y, $xflag);
}

sub EOR {
  $reg_a = $reg_a ^ &FetchMemA;
  &SetNeg($reg_a, $mflag);
  &SetZ($reg_a, $mflag);
}

sub INC {
  my $arg = &FetchMemA + 1;
  &StoreMem($arg, $mflag);
  &SetNeg($arg, $mflag);
  &SetZ($arg, $mflag);
}

sub INX {
  $reg_x += 1;
  &SetNeg($reg_x, $xflag);
  &SetZ($reg_x, $xflag);
}

sub INY {
  $reg_y += 1;
  &SetNeg($reg_y, $xflag);
  &SetZ($reg_y, $xflag);
}

sub JMP {
  $reg_pc = $effective_address;
}

sub JSR {
  my $blah = $reg_pc-1;
  &PushStack(int $blah / 256);
  &PushStack($blah - 256*int($blah/256));
  $reg_pc = $effective_address;
}

sub LDA {
  $reg_a = &FetchMemA;
  &SetNeg($reg_a, $mflag);
  &SetZ($reg_a, $mflag);
}

sub LDX {
  $reg_x = &FetchMemX;
  &SetNeg($reg_x, $xflag);
  &SetZ($reg_x, $xflag);
}

sub LDY {
  $reg_y = &FetchMemX;
  &SetNeg($reg_y, $xflag);
  &SetZ($reg_y, $xflag);
}

sub LSR {
  my $blah = &FetchMemA;
  my $carry = $blah & 0x01;
  $blah = $blah >> 1;
  if ($carry == 0) {
    &ClearFlag(0x01);
  } else {
    &SetFlag(0x01);
  }
  &SetNeg($blah, $mflag);
  &SetZ($blah, $mflag);
  &StoreMem($blah, $mflag);
}

sub MVN {
  &NotImplemented;
}

sub MVP {
  &NotImplemented;
}

sub NOP {
}

sub ORA {
  $reg_a = $reg_a | &FetchMemA;
  &SetNeg($reg_a, $mflag);
  &SetZ($reg_a, $mflag);
}

sub PEA {
  my $blah = $effective_address & 0xffff;
  &PushStack(int $blah / 256);
  &PushStack($blah & 255);
}

sub PEI {
  my $blah = $effective_address & 0xffff;
  &PushStack(int $blah / 256);
  &PushStack($blah & 255);
}

sub PER {
  my $blah = $effective_address & 0xffff;
  &PushStack(int $blah / 256);
  &PushStack($blah & 255);
}

sub PHA {
  if ($mflag == 0) {
    &PushStack(int $reg_a/256);
  }
  &PushStack($reg_a & 255);
}

sub PHB {
  &PushStack($reg_dbr);
}

sub PHD {
  &PushStack(int $reg_d/256);
  &PushStack($reg_d & 0xff);
}

sub PHK {
  &PushStack($reg_pbr);
}

sub PHP {
  &PushStack($reg_flags);
}

sub PHX {
  if ($xflag == 0) {
    &PushStack(int $reg_x/256);
  }
  &PushStack($reg_x & 255);
}

sub PHY {
  if ($xflag == 0) {
    &PushStack(int $reg_y/256);
  }
  &PushStack($reg_y & 255);
}

sub PLA {
  $reg_a = &PullStack;
  if ($mflag == 0) {
    $reg_a = $reg_a + 256*&PullStack;
  }
  &SetNeg($reg_a, $mflag);
  &SetZ($reg_a, $mflag);
}

sub PLB {
  $reg_dbr = &PullStack;
  &SetNeg($reg_dbr, 1);
  &SetZ($reg_dbr, 1);
}

sub PLD {
  $reg_d = &PullStack + 256*&PullStack;
  &SetNeg($reg_d, 0);
  &SetZ($reg_d, 0);
}

sub PLP {
  $reg_flags = &PullStack;
}

sub PLX {
  $reg_x = &PullStack;
  if ($xflag == 0) {
    $reg_x = $reg_x + 256*&PullStack;
  }
  &SetNeg($reg_x, $xflag);
  &SetZ($reg_x, $xflag);
}

sub PLY {
  $reg_y = &PullStack;
  if ($xflag == 0) {
    $reg_y = $reg_y + 256*&PullStack;
  }
  &SetNeg($reg_y, $xflag);
  &SetZ($reg_y, $xflag);
}

sub REP {
  &ClearFlag($effective_address);
}

sub ROL {
  my $blah = &FetchMemA;
  my $carry = $reg_flags & 0x01;
  $blah = ($blah << 1) | $carry;
  &SetCarry($blah);
  &SetNeg($blah, $mflag);
  &SetZ($blah, $mflag);
  &StoreMem($blah, $mflag);
}

sub ROR {
  my $blah = &FetchMemA;
  my $carry = $blah & 0x01;
  $blah = $blah >> 1;

  if (($reg_flags & 0x01) == 1) {
    if ($mflag == 1) {
      $blah = $blah | 0x80;
    } else {
      $blah = $blah | 0x8000;
    }
  }

  if ($carry == 0) {
    &ClearFlag(0x01);
  } else {
    &SetFlag(0x01);
  }
  &SetNeg($blah, $mflag);
  &SetZ($blah, $mflag);
  &StoreMem($blah, $mflag);
}

sub RTI {
  &NotImplemented;
}

sub RTL {
  $address[0] = &PullStack;
  $address[1] = &PullStack;
  $reg_pbr = &PullStack;
  $reg_pc = 256*$address[1] + $address[0] + 1;
}

sub RTS {
  $address[0] = &PullStack;
  $address[1] = &PullStack;
  $reg_pc = 256*$address[1] + $address[0] + 1;
}

sub SBC {
  my $argr = &FetchMemA;
  my $argl = $reg_a;
  my $carry = $reg_flags & 0x01;

  $reg_a = $argl - $argr - (1-$carry);
  if ($reg_a < 0) {
    $reg_a = &Neg16($reg_a);
    &ClearFlag(0x01);		# C clear
  } else {
    &SetFlag(0x01);
  }
  &SetNeg($reg_a, $mflag);
  &SetZ($reg_a, $mflag);
  &SetV($argl, $argr);
}

sub SEC {
  &SetFlag(0x01);
}

sub SED {
  &SetFlag(0x08);
}

sub SEI {
  &SetFlag(0x04);
}

sub SEP {
  &SetFlag($effective_address);
}

sub STA {
  &StoreMemA;
}

sub STP {
  &NotImplemented;
}

sub STX {
  &StoreMemX;
}

sub STY {
  &StoreMemY;
}

sub STZ {
  &StoreMem(0, $mflag);
}

sub TAX {
  $reg_x = $reg_a;
  &SetNeg($reg_x, $xflag);
  &SetZ($reg_x, $xflag);
}

sub TAY {
  $reg_y = $reg_a;
  &SetNeg($reg_y, $xflag);
  &SetZ($reg_y, $xflag);
}

sub TCD {
  $reg_d = $alo + 256*$ahi;
}

sub TCS {
  $reg_sp = $alo + 256*$ahi;
}

sub TDC {
  $alo = $reg_d & 0xff;
  $ahi = int $reg_d / 256;
  $reg_a = $alo + 256*$ahi;
}

sub TRB {
  my $blah = $reg_a & &FetchMemA;
  &StoreMem($blah, $mflag);
  &SetZ($blah, $mflag);
}

sub TSB {
  my $blah = $reg_a | &FetchMemA;
  &StoreMem($blah, $mflag);
  &SetZ($blah, $mflag);
}

sub TSC {
  $alo = $reg_sp & 0xff;
  $ahi = int $reg_sp / 256;
  $reg_a = $alo + 256*$ahi;
}

sub TSX {
  $xlo = $reg_sp & 0xff;
  $xhi = int $reg_sp / 256;
  $reg_x = $xlo + 256*$xhi;
}

sub TXA {
  $reg_a = $reg_x;		# High byte stuff works right?
  &SetNeg($reg_a, $mflag);
  &SetZ($reg_a, $mflag);
}

sub TXS {
  $reg_sp = $reg_x;
}

sub TXY {
  $reg_y = $reg_x;
  &SetNeg($reg_y, $xflag);
  &SetZ($reg_y, $xflag);
}

sub TYA {
  $reg_a = $reg_y;
  &SetNeg($reg_a, $mflag);
  &SetZ($reg_a, $mflag);
}

sub TYX {
  $reg_x = $reg_y;
  &SetNeg($reg_x, $xflag);
  &SetZ($reg_x, $xflag);
}

sub WAI {
  &NotImplemented;
}

sub WDM {
  &NotImplemented;
}

sub XBA {
  my $temp = $alo;
  $alo = $ahi;
  $ahi = $temp;
  $reg_a = $alo + 256*$ahi;
}

sub XCE {
  $emul = $reg_flags & 0x01;
  $reg_flags = ($reg_flags & 0xfe) | $reg_e;
  $reg_e = $emul;
}


##################################
#
# Utility routines
#
##################################

sub PushStack {
  $ram{0}[$reg_sp] = shift;	# Bank 0 always exists
  $reg_sp -= 1;
  if (($emul==1) && ($reg_sp < 0x0100)) {
    $reg_sp += 0x0100;
  }
}

sub PullStack {
  $reg_sp += 1;
  if (($emul==1) && ($reg_sp >= 0x0200)) {
    $reg_sp = 0x0100;
  }
  return $ram{0}[$reg_sp];	# Bank 0 always exists
}

#
# NOTE: On the C64, location 01 determines the memory configuration:
#
#   $A000-$BFFF  ROM/RAM
#   $D000-$DFFF  Hardware registers/CharROM/RAM
#   $E000-$FFFF  ROM/RAM
#
# A read from these locations will fetch ram/rom/regs depending on 01.
# A write to a ROM location always writes to the RAM underneath, whereas
# a write to a hardware register does not.
#
# See IsROM etc. for more info on the different mem configs.
#

sub FetchMemA {
  return &FetchMem($mflag);
}
sub FetchMemX {
  return &FetchMem($xflag);
}

sub FetchMem {
  my $flag = shift;		# $mflag or $xflag
  if ($mode eq "#") {
    return $address[0]+256*$address[1];
  }
  if ($mode eq "A") {		# e.g. ASL
    return $reg_a;
  }
  if (&IsReg($effective_address)) {	# Just return 0 for hardware regs
    return 0;
  }

  my $bank = int $effective_address/0x10000;
  my $addr = $effective_address - $bank*0x10000;
  &CheckRamBank($bank);
  if (&IsROM($effective_address)) {
    if ($flag == 0) {
      return $rom[$addr] + 256*$rom[$addr+1]; # Potential +1 bug
    } else {
      return $rom[$addr];
    }
  }

  if (($flag | $emul) == 0) {
    return $ram{$bank}[$addr] + 256*$ram{$bank}[$addr+1]; # Potential +1 bug
  } else {
    return $ram{$bank}[$addr];
  }
}

sub StoreMemA {
  return &StoreMem($reg_a, $mflag);
}
sub StoreMemX {
  return &StoreMem($reg_x, $xflag);
}
sub StoreMemY {
  return &StoreMem($reg_y, $xflag);
}
sub StoreMem {
  my $val = shift;
  my $flag = shift;		# $mflag or $xflag
  if ($flag == 1) {
    $val = $val & 0xffff;
  } else {
    $val = $val & 0xff;
  }

  if (&IsReg($effective_address)) {
    return;			# Just throw it away
  }
  if ($mode eq "A") {		# e.g. ASL
    $reg_a = $val;
    return;
  }
  my $bank = int $effective_address/0x10000;
  my $addr = $effective_address - $bank*0x10000;
  my $lo = $val & 0xff;
  my $hi = int $val/256;
  &CheckRamBank($bank);
  $ram{$bank}[$addr] = $lo;
  if (($flag | $emul) == 0) {
    $ram{$bank}[$addr+1] = $hi; # Potential +1 bug
  }
}

#
# Memory configurations are documented far above, at @a000/@d000/@e000 def.
#

sub IsROM {
  my $addr = shift;
  my $bits = $ram{0}[1] & 7;
  if ((0xa000 <= $addr) && ($addr <= 0xbfff)) {
    return ($a000[$bits] eq "rom");
  } elsif ((0xd000 <= $addr) && ($addr <= 0xdfff)) {
    return ($d000[$bits] eq "rom");
  } elsif ((0xe000 <= $addr) && ($addr <= 0xffff)) {
    return ($e000[$bits] eq "rom");
  } else {return 0;}
}

sub IsReg {
  my $addr = shift;
  my $bits = $ram{0}[1] & 7;
  if ((0xd000 <= $addr) && ($addr <= 0xdfff)) {
    return ($d000[$bits] eq "io");
  } else {return 0;}
}

#
# Set or clear flags.
#
sub SetCarry {		# .A or memory; always mflag
  my $arg = shift;
  my $max = 256;
  if ($mflag == 0) {
    $max = 65536;
  }
  if ($arg >= $max) {
    &SetFlag(0x01);
  } else {
    &ClearFlag(0x01);
  }
}

#
# According to Raeto West (Programming the Commodore 64), the V flag
# in the 6502 is (incorrectly) set as the inverse of the EOR of the
# sign bits.  I have not tested if this is the case with the 816 (and
# am not sufficiently motivated to test the 6502 either!) but assume it is
# and use it as such.
#

sub SetV {	# Always .A/mflag (ADC)
  my ($argl, $argr) = @_;
  my $arg = $argl ^ $argr;
  if (($emul | $mflag) == 1) {
    $arg = $arg & 0x80;
  } else {
    $arg = $arg & 0x8000;
  }
  if ($arg == 0) {
    &SetFlag(0x40);
  } else {
    &ClearFlag(0x40);
  }
}

sub SetNeg {
  my $arg = shift;
  my $flag = shift;
  my $mask = 0x80;
  if ($flag == 0) {
    $mask = 0x8000;
  }
  if (($arg & $mask) != 0) {
    &SetFlag(0x80);
  } else {
    &ClearFlag(0x80);
  }
}

sub SetZ {
  my $arg = shift;
  my $flag = shift;
  my $mask = 255;
  if ($flag == 0) {
    $mask = 65535;
  }
  if (($arg & $mask) == 0) {
    &SetFlag(0x02);
  } else {
    &ClearFlag(0x02);
  }
}

sub ClearFlag {
  my $mask = shift;
  $reg_flags = $reg_flags & ($mask ^ 0xff);
}

sub SetFlag {
  my $mask = shift;
  $reg_flags = $reg_flags | $mask;
}

sub BranchClear {
  my $mask = shift;

  if (($reg_flags & $mask) == 0) {
    $reg_pc = $effective_address;
  }
}

sub BranchSet {
  my $mask = shift;

  if (($reg_flags & $mask) != 0) {
    $reg_pc = $effective_address;
  }
}

sub CompareMem {
  my $reg = shift;
  my $flag = shift;

  my $arg = $reg - &FetchMem($flag);
  if ($arg < 0) {
    &ClearFlag(0x01);  # carry
    &SetFlag(0x80);    # neg
    &ClearFlag(0x02);  # Z
  } elsif ($arg > 0) {
    &SetFlag(0x01);  # carry
    &ClearFlag(0x80);    # neg
    &ClearFlag(0x02);  # Z
  } else {
    &SetFlag(0x01);  # carry
    &ClearFlag(0x80);    # neg
    &SetFlag(0x02);  # Z
  }
}

sub Neg16 {			# Just retain lower 16 bits if neg
  my $arg = shift;
  if ($arg < 0) {
    $arg = $arg & 0xffff;
  }
  return $arg;
}


sub Complement {
  my $arg = shift;
  if ($arg < 0) {
    $arg = (($arg ^ 0xffff) + 1) & 0xffff;
  }
}

sub CheckRamBank {
  my $bank = shift;
  if (!$ram{$bank}) {
    $ram{$bank} = [@rambank];
  }
}

sub CheckRamPBR {
  if (!$ram{$reg_pbr}) {
    $ram{$reg_pbr} = [@rambank];
  }
}

sub NotImplemented {
  die "$mnem (not implemented) at PC=", &ToHex16($reg_pc-1),
  	" .A=" , &ToHex16($reg_a) ,
	" .X=" , &ToHex16($reg_x) ,
	" .Y=" , &ToHex16($reg_y) ,
	" SR=" , &ToHex($reg_flags) ,
	" SP=" , &ToHex16($reg_sp) ,
	" e=" , $reg_e , "\n";
  #for my $j (0..3) {
  #  print &ToHex($ram{0}[$reg_sp+$j]), " ";
  #}
  #print "\n";
}
