/*
 * 65816.c
 *
 * This is a conversion of the 65816.pl program, which is too slow
 * for running programs.  Really it's just the core part of the code,
 * so that e.g. disassembly etc. can be done using perl or whatever.
 *
 * Stephen L. Judd
 * 3/19/05
 * last updated: 6/1/06
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "opdefs.h"
#include "opglobs.h"
#include "globals.h"
#include "prototypes.h"

#define DIR_SEPARATOR '/'

int main(int argc, char *argv[]) {

  long /*endaddr,*/ srcaddr;
  long org, codelen;
  char outfile[64] = "slangout.o";
  char srcfile[64] = "(no input file specified)";
  char userfile[64] = "--";
  char slangfile[64] = "SLANG1.4.O";
  char bindir[255] = "bin";
  char temp[64];
  int k,userflag=0,useraddr,tempa;

  for (k=1; k<argc; k++) {
    if (argv[k][0] == '-') {
      switch (argv[k][1]) {
	case 'o':
		strcpy(outfile, argv[k+1]);
      		k++;
		break;

	case 'x':
		strcpy(slangfile, argv[k+1]);
      		k++;
		break;

	case 'u':
		strcpy(userfile, argv[k+1]);
		userflag=1;
      		k++;
		break;

	case 'l':
		strcpy(bindir, argv[k+1]);
      		k++;
		break;

	case 'p':
		petsciiflag = 0;	// global
		break;

	case 'h':
		printf("Xlang 1.0 SLJ 11/11/05\n\n");
		printf("Usage: ");
		printf("xlang [-p] source.s [-x slangfile] [-o outfile]\n");
		printf("  -p  PETSCII source file (default=ASCII)\n");
		printf("  -x  Specify slang binary (e.g. -x SLANGB1.15.O)\n");
		printf("  -u  Specify userdef file\n");
		printf("  -l  Specify binary dir (e.g. -l ../bin)\n");
		printf("  -o  Specify output file\n");
		printf("  -h  Help\n");
		exit(0);

	default:
		printf("Unknown option %s\nxlang -h for help.\n", argv[k]);
		exit(0);
      }
    } else {
      strcpy(srcfile, argv[k]);
    }
  }

  Init65816();

  //Test #1
  //SimpleTest();

  //Test #2
  //LoadRAM("SPRITEDEMO.O", 0, 0);
  //IterateCode(loadaddr, 80);

  //printf("Loading ROM/RAM...\n");
  LoadROMBin(bindir,"rom-a000.prg", 0xa000, 0);
  LoadROMBin(bindir,"rom-e000.prg", 0xe000, 0);
  LoadROMBin(bindir,"CHARROM-D000.PRG", 0xd000, 0);
  LoadRAMBin(bindir,"ZP0000-03FF.PRG", 0, 0);
  LoadRAMBin(bindir,"SLANGDEF.O", 0, 0);
  if (userflag) {
    tempa=loadaddr;
    useraddr = ram[loadaddr+0x0a] + 256*ram[loadaddr+0x0b];
    LoadRAMBin(bindir,userfile,useraddr,0);
    if ((ram[useraddr]+256*ram[useraddr+1]) != useraddr) {
      printf("*** Incompatible userdef file -- recompile userdef.\n");
      return 0;
    } else {
      ram[tempa+0x0c] = 1;  /* userdef flag */
    }
  }
  LoadRAMBin(bindir,slangfile, 0, 0);

  srcaddr = 0x040000;
  if (petsciiflag == 1) {
    srcaddr = -srcaddr;		// Don't skip 1st two bytes of file
  }
  LoadRAM(srcfile, srcaddr, petsciiflag);
  srcaddr = labs(srcaddr);
  //effective_address = endloadaddr;
  //mode = _ABS_;		# For StoreMem
  //StoreMem(0, 1);	# Store a 0 at end of text
  ram[endloadaddr] = 0;
  //printf("Source code at %x-%x\n", startaddr, endaddr);
  printf("Compiling...\n");
  SetAXY(srcaddr);
  CallAsSub(0x0810);	// Init TEXTP
  SetAXY(endloadaddr);
  CallAsSub(0x0813);	// Init TEXTEND
  CallAsSub(0x0816);	// Assemble
  CallAsSub(0x0819);	// GetORG
  org = GetAXY();
  CallAsSub(0x081c);	// GetCODELEN; assembled code at $020000-$02CODELEN
  codelen = GetAXY();

  printf("\nORG=$%x, CODEEND=$%x\n", org, org+codelen);
  SaveObj(outfile, org, codelen);

  //&IterateCode(0x080d, 80);

  Cleanup65816();
  return 0;
}

/*
#
# Test code
#
*/

void SimpleTest() {
  long org=0x1000;
  long i;

  char code[21] = {0x20, 04, 0x10, 0x60, 0xa9, 01, 0x8d, 0x20, 0xd0, 0x60};
  //char code[21] = {0x18, 0xa9, 0xff, 0x69, 0x02, 0xb0, 0xfc, 0x0a, 0x00};
  //char code[21] = {0xa9, 0x00, 0x85, 0x02, 0xa9, 0x10, 0x85, 0x03, 0xa0, 0x00,
  //	   0xb1, 0x02, 0xd9, 0x00, 0x10, 0xc8, 0xc0, 0x04, 0xd0, 0xf6, 0x60};
  //unsigned char code[21] = {0xad, 0x00, 0xb0, 0xad, 0x20, 0xd0, 0xad, 0xd2, 0xff};

  mode = _ABS_;		// For StoreMem
  for (i=0; i<21; i++) {
    effective_address = org+i;
    StoreMem(code[i], 1);
  }
  ram[0xb000] = 0x3d;
  ram[0xd020] = 0xff;
  ram[0xffd2] = 0xc0;

  IterateCode(0x1000, 40);
  //CallAsSub(0x1000);
}



void IterateCode(long addr, long iter) {
  long i;

  SetPC(addr);
  for (i=0; i<iter; i++) {
    ExOp();
    PrintDisassembly();
  }
}

void PrintDisassembly() {
    int i;
    char txt[32];

    ToHex16(txt, oppc);
    printf("%s> ", txt);
    for (i=0; i<4; i++) {
      if (i < numbytesread) {
        ToHex(txt, bytesread[i]);
        printf("%s ", txt);
      } else {
        printf("   ");
      }
    }

    DisasOp(txt);
    printf("%-11s", txt);
    ToHex16(txt, reg_a);
    printf(" .A=%s", txt);
    ToHex16(txt, reg_x);
    printf(" .X=%s", txt);
    ToHex16(txt, reg_y);
    printf(" .Y=%s", txt);
    ToHex(txt, reg_flags);
    printf(" SR=%s", txt);
    ToHex16(txt, reg_sp);
    printf(" SP=%s>", txt);
    for (i=0; i<4; i++) {
      ToHex(txt, ram[reg_sp+i]);
      printf("%s ", txt);
    }
    printf("\n");
}

void DumpMem(long start, long end) {
  int count=0;
  long i;
  char txt[5], txt2[17];

  for (i=start; i<=end; i++) {
    if (count == 0) {
      printf("%x>", i);
    }
    ToHex(txt, ram[i]);
    printf("%s ", txt);
    if ((ram[i] & 0x7f) >= 32) {
      txt2[count] = ram[i] & 0x7f;
    } else {
      txt2[count] = 46;	//dot
    }
    count++;
    if (count > 15) {
      txt2[count] = 0;
      printf("%s\n", txt2);
      count = 0;
    }
  }
  if (count != 0) {
    while (count < 16) {
      printf("   ");
      txt2[count] = 32;	//space
      count++;
    }
    txt2[count] = 0;
    printf("%s\n", txt2);
  }
}

/*
#
# CallAsSub is the main code execution routine.  It pretends to call
# the address parameter as a JSR, by setting PC=$FFFE and waiting for
# an RTS to set it back.
#
# This code also intercepts some of the kernal routines, including
# FFD2, OPEN, LOAD, etc.
#
*/
void CallAsSub(long addr) {
  long debug_iter=0;

  effective_address = addr;
  SetPC(0xffff);
  CallOp(JSR);
  while (reg_pc < 0xffff) {
//    debug_iter++;
    ExOp();
//    PrintDisassembly();
    if (reg_pc == 0xffd2) {	// CHROUT
      Kernal_CHROUT();
    } else if (reg_pc == 0xffc0) {
      Kernal_OPEN();
    } else if (reg_pc == 0xffd5) {
      Kernal_LOAD();
    } else if (reg_pc == 0xffcf) {
      Kernal_CHRIN();
    } else if (reg_pc == 0xffe4) {
      Kernal_CHRIN();		// Only might be used for reading files
    }
//    if ((debug_iter > 4000000)) { // && (ram[0x0400ae] == 0)){
//      PrintDisassembly();
//      printf("bug: debug_iter=%ld\n", debug_iter);
//      DumpMem(0x0400ad, 0x0400bc);
//    }

/***
    if (reg_pc == 0xbcf3) {  // float conversion bug
      PrintDisassembly();
      DumpMem(0x61, 0x66);
      DumpMem(0x79, 0x7b);
      DumpMem(0x0200, 0x0208);
debug_iter = -1;
    }
if (debug_iter < 0) {
  PrintDisassembly();
  DumpMem(0x61,0x66);
  DumpMem(0x69,0x6e);
}
***/

  }
//      printf("bug: debug_iter=%ld\n", debug_iter);
}

/*
 * Kernal routines.  Note that only the parts of these routines that
 * are used by Slang have been emulated, as opposed to emulating the
 * full routine functionality.
 */

void Kernal_CHROUT() {
      char chr[2];

      chr[1]=0;
      chr[0] = (char)(reg_a & 0x7f);
      if (reg_a < 32) {	// Could sorta use tr, but...
	chr[0] = '\n';
      } else if ((reg_a > 64) && (reg_a < 92)) {
	chr[0] = (char)(reg_a + 32);
      } else if ((reg_a > 96) && (reg_a < 128)) {
	chr[0] = (char)(reg_a - 32);
      } else if ((reg_a > 192) && (reg_a < 224)) {
	chr[0] = (char)(reg_a - 128);
      }
      printf("%s", chr);
      CallOp(RTS);
}

void Kernal_OPEN() {
  char fname[64];

  if (ram[0xba] >= 8) {		// Kernal uses e.g. device 3
    GetKernalFilename(fname);
    OPENFH = fopen(fname, "rb");
    if (!OPENFH) {
      printf("\nCan't open file '%s' -- exiting...\n", fname);
      exit(0);
    }
    ram[0x90] = 0;
    CallOp(RTS);
  }
}

void Kernal_LOAD() {		// Assume binary data?
  char fname[64];
  unsigned char sa;
  long faddr;

  GetKernalFilename(fname);
  sa = ram[0xb9];
  if (sa == 255) {		// 255 => PPPatch 24-bit addr
    faddr = reg_x + 256*(reg_y + 256*ram[0xb0]);
    LoadRAM(fname, faddr, 2);
  } else if (sa == 0) {
    faddr = reg_x + 256*reg_y;
    LoadRAM(fname, faddr, 2);
  } else {
    LoadRAM(fname, 0, 2);
  }
  CallOp(CLC);			//success; failure exits below
  CallOp(RTS);
}

void Kernal_CHRIN() {		// used by PUT(?), so petsciiflag is used
  char byte;

  if (ram[0xba] >= 8) {
    if (fread(&byte, 1, 1, OPENFH)) {
      byte = ToPetscii(byte);
      reg_a = byte;
    } else {
      reg_a = 0x0d;
      ram[0x90] = 64;
    }
    CallOp(RTS);
  }
}

void GetKernalFilename(char *fname) {
  unsigned char flen, k;
  long faddr;

  flen = ram[0xb7];
  if (flen > 63) {
    printf("Is that long of a filename REALLY necessary?\n");
    exit(0);
  }
  faddr = ram[0xbb] + 256*ram[0xbc];
  for (k=0; k<flen; k++) {
    fname[k] = FromPetscii(ram[faddr+k]);
  }
  fname[flen] = 0;
}

unsigned char ToPetscii(unsigned char byte) {
      if (petsciiflag == 1) {
        if (byte < 32) {
	  if (byte == 13) {	// Convert CR LF -> SPACE LF
	    byte = 32;
	  } else if (byte == 9) { // tab -> space
	    byte = 32;
	  } else {
  	    byte = 13;		// Convert all to CR
	  }
        }
	if ((byte > 64) && (byte < 91)) {	// Poor man's conversion
	  byte += 32;
	} else if ((byte > 96) && (byte < 128)) {
	  byte -= 32;
	}
      }
  return byte;
}

unsigned char FromPetscii(unsigned char byte) {
      if (petsciiflag == 1) {
	if ((byte > 64) && (byte < 95)) {
	  byte += 32;
	} else if ((byte > 96) && (byte < 123)) {
	  byte -= 32;
	}
      }
  return byte;
}

/*
#
# The Sirius interface uses .AXY to set TEXTP and TEXTEND.
#
*/

void SetAXY(long arg) {
  reg_a = arg & 0xff;
  reg_x = (arg >> 8) & 0xff;
  reg_y = (arg >> 16) & 0xff;
}

long GetAXY() {
  return reg_a + 256*(reg_x + 256*reg_y);
}

/*
#
# Load files to ram/rom
#
# Input:
#   filename
#   load address   x -> load to address x, skip 1st two bytes of file
#		  -x -> load to x, do not skip 1st two bytes
#		   0 -> use file load address (1st two bytes of file)
#   petscii flag (optional) - =1 means strip LF/convert to CR.
#
# Returns:
#   Load start address, load end address+1
#   $ae set to end addr+1
#
*/

void LoadRAMBin(char *bindir, char *fname, long addr, int petscii) {
  char temp[255];
  sprintf(temp, "%s%c%s", bindir, DIR_SEPARATOR, fname);
  LoadRAM(temp, addr, petscii);
}

void LoadRAM(char *fname, long addr, int petscii) {
  LoadMem(ram, fname, addr, petscii);
  ram[0xae] = endloadaddr & 0xff; // Set ZP pointers correctly, for PUT
  ram[0xaf] = (endloadaddr >> 8) & 0xff;
  ram[0xb0] = (endloadaddr >> 16) & 0xff;
  //DumpMem(0xae, 0xb0);
}

void LoadROMBin(char *bindir, char *fname, long addr, int petscii) {
  char temp[255];
  sprintf(temp, "%s%c%s", bindir, DIR_SEPARATOR, fname);
  LoadROM(temp, addr, petscii);
}

void LoadROM(char *fname, long addr, int petscii) {
  LoadMem(rom, fname, addr, petscii);
}

void LoadMem(char *mem, char *fname, long addr, int petscii) {
  unsigned char byte;
  long faddr, count;
  char temp[20];
  FILE *fh;

  /* detect if it's really petscii first */
  if(petscii==2)
  {
     int p=0;
     petscii=1;
     fh=fopen(fname,"rb");
     while(p!=EOF)
     {
        p=fgetc(fh);
        if(p>=128)
        {
           petscii=0;
           break;
        }
     }
     fclose(fh);
  }

  fh = fopen(fname, "rb");
  printf("Loading (%s) %s...", petscii?"ascii":"petscii", fname);
  if (!fh) {
    printf("\nCan't open file %s -- exiting...\n", fname);
    exit(0);
  }

  if (addr >= 0) {
    fread(&byte, 1, 1, fh);
    faddr = byte;
    fread(&byte, 1, 1, fh);
    faddr = faddr + 256*byte;
    if (addr == 0) {
      addr = faddr;
    }
  } else {
    addr = -addr;
  }
  ToHex24(temp, addr);
  printf(" $%s-", temp);
  loadaddr = addr;		// start address (global)

  count = 0;
  if (petscii != 1) {
    while (!feof(fh)) {
      count = fread(&mem[addr], 1, 4096, fh);
      addr+=count;
    }
  } else {
    while (!feof(fh)) {
      if (fread(&byte, 1, 1, fh)) {
        mem[addr] = ToPetscii(byte);
        addr++;
      }
    }
  }
  fclose(fh);
  ToHex24(temp, addr);
  printf("$%s\n", temp);
  endloadaddr = addr;		// End address +1 (global)
}

/*
#
# Save the object code at $020000, $len bytes total
#
*/
void SaveObj(char *fname, long org, long len) {

  unsigned char byte;
  /*long faddr, k;*/
  FILE *fh;

  fh = fopen(fname, "wb");
  printf("Saving %s... $%x-$%x\n", fname, org, org+len);
  if (!fh) {
    printf("\nCan't open file %s -- exiting...\n", fname);
    exit(0);
  }

  byte = (unsigned char)(org & 0xff);
  fwrite(&byte, 1, 1, fh);
  byte = (unsigned char)(org >> 8) & 0xff;
  fwrite(&byte, 1, 1, fh);

  fwrite(&ram[0x020000], 1, len, fh);
  fclose(fh);
}

/*
#
# Init
#
*/
void Init65816() {
  ram = (char *) calloc(0x050000, 1);
  ram[1] = 0x00;	// Force default $01 = all RAM.
  rom = (char *) calloc(0x010000, 1);
}

void Cleanup65816() {
  free(ram);
  free(rom);
}

/*
#
# Execute an opcode: read opcode and address, call routine.
#
*/
void ExOp() {
  oppc = reg_pc;
  ParseOp();
  CallOp(mnem);
  FixRegs();
}

void ParseOp() {
  int i;
  int plus;

  address[0] = 0;
  address[1] = 0;
  address[2] = reg_dbr;
  //@address = (0, 0, $reg_dbr);
  numbytesread = 0;
  //@bytesread = ();

  opcode = ReadByte();
  mnem = oplist[2*opcode];
  mode = oplist[2*opcode + 1];
  bytesread[numbytesread] = opcode;
  numbytesread++;
  //if (!exists $modelen{$mode}) {die "Unknown addressing mode: $mnem $mode\n";}
  plus = CheckModeLen();
  for (i=0; i < modelen[mode]+plus; i++) {
    address[i] = ReadByte();
    bytesread[numbytesread] = address[i];
    numbytesread++;
  }
  FixAddress();
}

/*
#
# If #immed mode, and in native mode, and register flag is set
# appropriately, then need to read an extra byte.
#
*/

int CheckModeLen() {
  int val;

  if (emul == 1) {
    return 0;
  }
  if (mode != _IMMED_) {
    return 0;
  }

  switch (mnem) {
    case ADC:
	    val = 1-mflag;
    case AND:
	    val = 1-mflag;
    case BIT:
	    val = 1-mflag;
    case CMP:
	    val = 1-mflag;
    case EOR:
	    val = 1-mflag;
    case LDA:
	    val = 1-mflag;
    case ORA:
	    val = 1-mflag;
    case SBC:
	    val = 1-mflag;

    case CPX:
	    val = 1-xflag;
    case CPY:
	    val = 1-xflag;
    case LDX:
	    val = 1-xflag;
    case LDY:
	    val = 1-xflag;

    default:
	    val = 0;
  }
  return val;
}

/*
#
# Read a byte using PBR and PC
# On exit: PC -> PC+1
#
*/

unsigned char ReadByte() {
  long addr;

  //CheckRamPBR();
  addr = (reg_pbr << 16) + reg_pc;
  reg_pc = reg_pc + 1;

  if (IsROM(addr)) {
    return rom[addr];
  } else {
    return ram[addr];
  }
}

/*
#
# Fix up an address based on the addressing mode.  The result is a
# 24-bit address.  Note that this is required to handle passing bank
# boundaries correctly (RAM is stored as a hash, not a fixed+flat array).
#
*/

void FixAddress() {
  unsigned int zp;

  effective_address = address[0] + 256*(address[1] + 256*address[2]);
  zp = address[0] + 256*address[1] + reg_d; // Bank 0, for zp guys

  switch (mode) {
    case _IMMED_:
	break;

    case _IABS_:
    	effective_address = GetIndirect16(0);
	break;

    case _IABSX_:
    	effective_address = GetIndirect16(reg_x);
	break;

    case _IDP_:
    	effective_address = GetZPPtr16(0);
	break;

    case _IDPY_:
    	effective_address = GetZPPtr16(0) + reg_y;
	break;

    case _STACKRELY_:
    	effective_address = GetZPPtr16(reg_sp) + reg_y;
	break;

    case _IDPX_:
    	effective_address = GetZPPtr16(reg_x);
	break;

    case _ACCUM_:
	break;

    case _LIDP_:
    	effective_address = GetZPPtr24();
	break;

    case _LIDPY_:
    	effective_address = GetZPPtr24() + reg_y;
	break;

    case _ABS_:
	break;

    case _ABSX_:
    	effective_address += reg_x;
	break;

    case _ABSY_:
    	effective_address += reg_y;
	break;

    case _LONGABS_:
	break;

    case _LONGABSX_:
    	effective_address += reg_x;
	break;

    case _DP_:
    	effective_address = zp;
	break;

    case _STACKREL_:
    	effective_address = (zp + reg_sp) & 0xffff; // Potential +1 bug
	break;

    case _DPX_:
    	effective_address = zp + reg_x;
	if (emul == 1) {	// 6502 is 8-bit effective address
	  effective_address = effective_address & 0xff;
	}
	break;

    case _DPY_:
    	effective_address = zp + reg_y;
	if (emul == 1) {	// 6502 is 8-bit effective address
	  effective_address = effective_address & 0xff;
	}
	break;

    case _IMPLIED_:
	break;

    case _REL_:
      if (address[0] >= 128) {
	effective_address -= 256;
      }
      effective_address = reg_pc + effective_address;
	break;

    case _LONGREL_:
      if (address[1] >= 128) {
	effective_address -= 0x10000;
      }
      effective_address = reg_pc + effective_address;
	break;

    case _STACK_:
	break;

    case _XYC_:
      	printf("Unimplemented mode XYC in FixAddress\n");
	break;

    default:
	printf("Unknown mode in FixAddress?!\n");
  }

}

long GetIndirect16(long offset) {
  long addr, addr2;
  unsigned char lo, hi;

  addr = (effective_address + offset) & 0xffff;
  addr2 = (effective_address + offset + 1) & 0xffff;
  lo = ram[addr];
  hi = ram[addr2];
  return lo + 256*(hi + 256*reg_dbr);
}

long GetZPPtr16(long offset) {
  long addr, addr2;
  unsigned char lo, hi;

  addr = (effective_address + reg_d + offset) & 0xffff;
  addr2 = (effective_address + reg_d + offset + 1) & 0xffff;
  lo = ram[addr];
  hi = ram[addr2];
  return lo + 256*(hi + 256*reg_dbr);
}

long GetZPPtr24() {
  long addr, addr2, addr3;
  unsigned char lo, hi, hi24;

  addr = (effective_address + reg_d) & 0xffff;
  addr2 = (effective_address + reg_d + 1) & 0xffff;
  addr3 = (effective_address + reg_d + 2) & 0xffff;
  lo = ram[addr];
  hi = ram[addr2];
  hi24 = ram[addr3];
  return lo + 256*(hi + 256*hi24);
}

/*
#
# Fix up the register bitness, set flag shortcuts
#
*/

void FixRegs() {
  emul = reg_e;
  cflag = reg_flags & 01;
  if (emul == 1) {
    mflag=1;
    xflag=1;
    reg_sp = 0x0100 + (reg_sp & 0xff);
  } else {
    mflag = (reg_flags & 0x20) / 0x20;
    xflag = (reg_flags & 0x10) / 0x10;
  }

  if ((mflag | emul) == 1) {
    reg_a = reg_a & 255;
    alo = (unsigned char)reg_a;
  } else {
    alo = (unsigned char)(reg_a & 255) ;
    ahi = (unsigned char)(reg_a >> 8);
  }

  if ((xflag | emul) == 1) {	// note: in emul mode, xhi = 0
    reg_x = reg_x & 255;
    xlo = (unsigned char)reg_x;
    xhi = 0;
    reg_y = (unsigned char)(reg_y & 255);
    ylo = (unsigned char)reg_y;
    yhi = 0;
  } else {
    xlo = (unsigned char)(reg_x & 255);
    xhi = (unsigned char)(reg_x/256);
    ylo = (unsigned char)(reg_y & 255);
    yhi = (unsigned char)(reg_y/256);
  }
}

/*
#
# Set the program counter + pbr
#
*/

void SetPC(long val) {
  reg_pbr = val >> 16;
  reg_pc = val & 0xffff;
}

/*
##################################
#
# Disassembly routines
#
##################################
*/

/*
#
# Disassemble current opcode
# Returns string [op arg] and array of bytes
#
*/
void DisasOp(char *dest) {
  int i;
  char str[20];
  char arg[20] = "$";
  char temp[20];			// C strings... are... crap?

  //strcpy(str, opcodes[mnem]);
  //strcat(str, " ");
  str[0]=0;
  sprintf(str, "%s ", opcodes[mnem]);

  strcpy(arg, "$");
  for (i=numbytesread-1; i>0; i--) {
    ToHex(temp, bytesread[i]);
    strcat(arg, temp);
  }

  switch (mode) {
    case _IMMED_:
	    //strcpy(temp, "#");
	    //strcat(temp, arg);
	    sprintf(temp, "#%s", arg);
	    strcpy(arg, temp);
	    break;

    case _IABS_:
	    //strcpy(temp, "(");
	    //strcat(temp, arg);
	    //strcat(temp, ")");
	    sprintf(temp, "(%s)", arg);
	    strcpy(arg, temp);
	    break;

    case _IABSX_:
	    //strcpy(temp, "(");
	    //strcat(temp, arg);
	    //strcat(temp, ",x");
	    //strcpy(arg, temp);
	    sprintf(arg, "(%s),x", arg);
	    break;

    case _IDP_:
	    strcpy(temp, "(");
	    strcat(temp, arg);
	    strcat(temp, ")");
	    strcpy(arg, temp);
	    break;

    case _IDPY_:
	    strcpy(temp, "(");
	    strcat(temp, arg);
	    strcat(temp, "),y");
	    strcpy(arg, temp);
	    break;

    case _STACKRELY_:
	    strcpy(temp, "(");
	    strcat(temp, arg);
	    strcat(temp, ",s),y");
	    strcpy(arg, temp);
	    break;

    case _IDPX_:
	    strcpy(temp, "(");
	    strcat(temp, arg);
	    strcat(temp, ",x)");
	    strcpy(arg, temp);
	    break;

    case _ACCUM_:
	    strcpy(arg, "");
	    break;

    case _LIDP_:
	    strcpy(temp, "[");
	    strcat(temp, arg);
	    strcat(temp, "]");
	    strcpy(arg, temp);
	    break;

    case _LIDPY_:
	    strcpy(temp, "[");
	    strcat(temp, arg);
	    strcat(temp, "],y");
	    strcpy(arg, temp);
	    break;

    case _ABS_:
	    break;

    case _ABSX_:
	    strcat(arg, ",x");
	    break;

    case _ABSY_:
	    strcat(arg, ",y");
	    break;

    case _LONGABS_:
	    break;

    case _LONGABSX_:
	    strcat(arg, ",x");
	    break;

    case _DP_:
	    break;

    case _STACKREL_:
	    strcat(arg, ",s");
	    break;

    case _DPX_:
	    strcat(arg, ",x");
	    break;

    case _DPY_:
	    strcat(arg, ",y");
	    break;

    case _IMPLIED_:
	    strcpy(arg, "");
	    break;

    case _REL_:
	    ToHex16(temp, effective_address);
	    //strcat(arg, temp);
	    sprintf(arg, "$%s", temp);
	    break;

    case _LONGREL_:
	    ToHex16(temp, effective_address);
	    strcat(arg, temp);
	    break;

    case _STACK_:
	    strcpy(arg, "");
	    break;

    case _XYC_:
	    strcpy(arg, "XYC (fixme!)");
	    break;

    default:
	    printf("Whoah, weird error in subroutine DisasOp!\n");
  }

  strcat(str, arg);
  strcpy(dest, str);
}

/*
#
# Convert a number to a 2-digit hex value, append to dest.
#
*/
void ToHex(char *dest, long val) {
  char str[3] = "abc";
  char hex[17] = "0123456789abcdef";
  char hi = (char)(val >> 4) & 0x0f;
  char lo = (char)val & 0x0f;
  str[0] = hex[hi];
  str[1] = hex[lo];
  str[2] = 0;
  //return str;
  strcpy(dest, str);
}

void ToHex16(char *dest, long val) {
  char str[3];
  ToHex(dest, val >> 8);
  ToHex(str, val & 0xff);
  strcat(dest, str);
}

void ToHex24(char *dest, long val) {
  char str[3];
  ToHex(dest, val >> 16);
  ToHex(str, val >> 8);
  strcat(dest, str);
  ToHex(str, val & 0xff);
  strcat(dest, str);
}

/*
##################################
#
# Opcodes
#
##################################
*/

void CallOp(unsigned char instruction) {

  unsigned int argl, argr, blah, mem, hi, /*lo,*/ carry;

  switch (instruction) {

    case ADC :
  	argr = FetchMemA();
  	argl = reg_a;
  	/*reg_a += argr + cflag;*/ /*iAN*/
	if (reg_flags & 0x08) {	// BCD mode... bleah... SBC not fixed...
	  if ((reg_a & 0x0f) > 9) {
	    reg_a+= 0x06;
	  }
	  if (((reg_a >> 4) & 0x0f) > 9) {
	    reg_a+= 0x60;
	  }
	  if (((reg_a >> 8) & 0x0f) > 9) {
	    reg_a+= 0x600;
	  }
	  if (((reg_a >> 12) & 0x0f) > 9) {
	    reg_a+= 0x6000;
	  }
	}
	reg_a += argr + cflag; /*iAN*/
  	SetCarry(reg_a);
  	SetV(argl, argr);
  	SetNeg(reg_a, mflag);
  	SetZ(reg_a, mflag);
	break;

    case AND :
  	reg_a = reg_a & FetchMemA();
  	SetNeg(reg_a, mflag);
  	SetZ(reg_a, mflag);
	break;

    case ASL :
  	blah = FetchMemA() << 1;
  	SetCarry(blah);
  	SetNeg(blah, mflag);
  	SetZ(blah, mflag);
  	StoreMem(blah, mflag);
	break;

    case BCC :
  	BranchClear(0x01);
	break;

    case BCS :
  	BranchSet(0x01);
	break;

    case BEQ :
  	BranchSet(0x02);
	break;

    case BIT :
  	mem = FetchMemA();
  	argl = reg_a & mem;
  	hi = mem & 0xc0;		// M7 and M6
  	if (mflag == 0) {
  	  hi = mem & 0xc000;
  	}

  	SetZ(argl, mflag);
  	if (mode != _IMMED_ ) {		// M7->N, M6->V
  	  reg_flags = ((reg_flags & 0x3f) | hi);
 	}
	break;

    case BMI :
  	BranchSet(0x80);
	break;

    case BNE :
  	BranchClear(0x02);
	break;

    case BPL :
  	BranchClear(0x80);
	break;

    case BRA :
  	reg_pc = effective_address;
	break;

    case BRK :
  	NotImplemented();
	break;

    case BRL :
  	reg_pc = effective_address;
	break;

    case BVC :
  	BranchClear(0x40);
	break;

    case BVS :
  	BranchSet(0x40);
	break;

    case CLC :
  	ClearFlag(0x01);
	break;

    case CLD :
  	ClearFlag(0x08);
	break;

    case CLI :
  	ClearFlag(0x04);
	break;

    case CLV :
  	ClearFlag(0x40);
	break;

    case CMP :
  	CompareMem(reg_a, mflag);
	break;

    case COP :
  	NotImplemented();
	break;

    case CPX :
  	CompareMem(reg_x, xflag);
	break;

    case CPY :
  	CompareMem(reg_y, xflag);
	break;

    case DEC :
  	argl = Neg16(FetchMemA() - 1);
  	StoreMem(argl, mflag);
  	SetNeg(argl, mflag);
  	SetZ(argl, mflag);
	break;

    case DEX :
  	reg_x = Neg16(reg_x - 1);
  	SetNeg(reg_x, xflag);
  	SetZ(reg_x, xflag);
	break;

    case DEY :
  	reg_y = Neg16(reg_y - 1);
  	SetNeg(reg_y, xflag);
  	SetZ(reg_y, xflag);
	break;

    case EOR :
  	reg_a = reg_a ^ FetchMemA();
  	SetNeg(reg_a, mflag);
  	SetZ(reg_a, mflag);
	break;

    case INC :
  	argl = FetchMemA() + 1;
  	StoreMem(argl, mflag);
  	SetNeg(argl, mflag);
  	SetZ(argl, mflag);
	break;

    case INX :
  	reg_x += 1;
  	SetNeg(reg_x, xflag);
  	SetZ(reg_x, xflag);
	break;

    case INY :
  	reg_y += 1;
  	SetNeg(reg_y, xflag);
  	SetZ(reg_y, xflag);
	break;

    case JMP :
  	reg_pc = effective_address;
	break;

    case JSR :
  	blah = reg_pc-1;
  	PushStack(blah >> 8);
  	PushStack(blah & 0xff);
  	reg_pc = effective_address;
	break;

    case LDA :
  	reg_a = FetchMemA();
  	SetNeg(reg_a, mflag);
  	SetZ(reg_a, mflag);
	break;

    case LDX :
  	reg_x = FetchMemX();
  	SetNeg(reg_x, xflag);
  	SetZ(reg_x, xflag);
	break;

    case LDY :
  	reg_y = FetchMemX();
  	SetNeg(reg_y, xflag);
  	SetZ(reg_y, xflag);
	break;

    case LSR :
  	blah = FetchMemA();
  	carry = blah & 0x01;
  	blah = blah >> 1;
  	if (carry == 0) {
  	  ClearFlag(0x01);
  	} else {
  	  SetFlag(0x01);
  	}
  	SetNeg(blah, mflag);
  	SetZ(blah, mflag);
  	StoreMem(blah, mflag);
	break;

    case MVN :
  	NotImplemented();
	break;

    case MVP :
  	NotImplemented();
	break;

    case NOP :
	break;

    case ORA :
  	reg_a = reg_a | FetchMemA();
  	SetNeg(reg_a, mflag);
  	SetZ(reg_a, mflag);
	break;

    case PEA :
  	blah = effective_address & 0xffff;
  	PushStack(blah >> 8);
  	PushStack(blah & 255);
	break;

    case PEI :
  	blah = effective_address & 0xffff;
  	PushStack(blah >> 8);
  	PushStack(blah & 255);
	break;

    case PER :
  	blah = effective_address & 0xffff;
  	PushStack(blah >> 8);
  	PushStack(blah & 255);
	break;

    case PHA :
  	if (mflag == 0) {
  	  PushStack(reg_a >> 8);
  	}
  	PushStack(reg_a & 255);
	break;

    case PHB :
  	PushStack(reg_dbr);
	break;

    case PHD :
  	PushStack(reg_d >> 8);
  	PushStack(reg_d & 0xff);
	break;

    case PHK :
  	PushStack(reg_pbr);
	break;

    case PHP :
  	PushStack(reg_flags);
	break;

    case PHX :
  	if (xflag == 0) {
  	  PushStack(reg_x >> 8);
  	}
  	PushStack(reg_x & 255);
	break;

    case PHY :
  	if (xflag == 0) {
  	  PushStack(reg_y >> 8);
  	}
  	PushStack(reg_y & 255);
	break;

    case PLA :
  	reg_a = PullStack();
  	if (mflag == 0) {
  	  reg_a = reg_a + 256*PullStack();
  	}
  	SetNeg(reg_a, mflag);
  	SetZ(reg_a, mflag);
	break;

    case PLB :
  	reg_dbr = PullStack();
  	SetNeg(reg_dbr, 1);
  	SetZ(reg_dbr, 1);
	break;

    case PLD :
  	reg_d = PullStack() + 256*PullStack();
  	SetNeg(reg_d, 0);
  	SetZ(reg_d, 0);
	break;

    case PLP :
  	reg_flags = PullStack();
	break;

    case PLX :
  	reg_x = PullStack();
  	if (xflag == 0) {
  	  reg_x = reg_x + 256*PullStack();
  	}
  	SetNeg(reg_x, xflag);
  	SetZ(reg_x, xflag);
	break;

    case PLY :
  	reg_y = PullStack();
  	if (xflag == 0) {
  	  reg_y = reg_y + 256*PullStack();
  	}
  	SetNeg(reg_y, xflag);
  	SetZ(reg_y, xflag);
	break;

    case REP :
  	ClearFlag(effective_address);
	break;

    case ROL :
  	blah = FetchMemA();
  	carry = reg_flags & 0x01;
  	blah = (blah << 1) | carry;
  	SetCarry(blah);
  	SetNeg(blah, mflag);
  	SetZ(blah, mflag);
  	StoreMem(blah, mflag);
	break;

    case ROR :
  	blah = FetchMemA();
  	carry = blah & 0x01;
  	blah = blah >> 1;

  	if ((reg_flags & 0x01) == 1) {
  	  if (mflag == 1) {
  	    blah = blah | 0x80;
  	  } else {
  	    blah = blah | 0x8000;
  	  }
  	}

  	if (carry == 0) {
  	  ClearFlag(0x01);
  	} else {
  	  SetFlag(0x01);
  	}
  	SetNeg(blah, mflag);
  	SetZ(blah, mflag);
  	StoreMem(blah, mflag);
	break;

    case RTI :
  	NotImplemented();
	break;

    case RTL :
  	address[0] = PullStack();
  	address[1] = PullStack();
  	reg_pbr = PullStack();
  	reg_pc = 256*address[1] + address[0] + 1;
	break;

    case RTS :
  	address[0] = PullStack();
  	address[1] = PullStack();
  	reg_pc = 256*address[1] + address[0] + 1;
	break;

    case SBC :
  	argr = FetchMemA();
  	argl = reg_a;
  	carry = reg_flags & 0x01;

  	reg_a = argl - argr - (1-carry);
  	if (reg_a < 0) {
  	  reg_a = Neg16(reg_a);
  	  ClearFlag(0x01);		// C clear
  	} else {
  	  SetFlag(0x01);
  	}
  	SetNeg(reg_a, mflag);
  	SetZ(reg_a, mflag);
  	SetV(argl, argr);
	break;

    case SEC :
  	SetFlag(0x01);
	break;

    case SED :
  	SetFlag(0x08);
	break;

    case SEI :
  	SetFlag(0x04);
	break;

    case SEP :
  	SetFlag(effective_address);
	break;

    case STA :
  	StoreMemA();
	break;

    case STP :
  	NotImplemented();
	break;

    case STX :
  	StoreMemX();
	break;

    case STY :
  	StoreMemY();
	break;

    case STZ :
  	StoreMem(0, mflag);
	break;

    case TAX :
  	reg_x = reg_a;
  	SetNeg(reg_x, xflag);
  	SetZ(reg_x, xflag);
	break;

    case TAY :
  	reg_y = reg_a;
  	SetNeg(reg_y, xflag);
  	SetZ(reg_y, xflag);
	break;

    case TCD :
  	reg_d = alo + 256*ahi;
	break;

    case TCS :
  	reg_sp = alo + 256*ahi;
	break;

    case TDC :
  	alo = reg_d & 0xff;
  	ahi = reg_d >> 8;
  	reg_a = alo + 256*ahi;
	break;

    case TRB :
  	blah = reg_a & FetchMemA();
  	StoreMem(blah, mflag);
  	SetZ(blah, mflag);
	break;

    case TSB :
  	blah = reg_a | FetchMemA();
  	StoreMem(blah, mflag);
  	SetZ(blah, mflag);
	break;

    case TSC :
  	alo = reg_sp & 0xff;
  	ahi = reg_sp >> 8;
  	reg_a = alo + 256*ahi;
	break;

    case TSX :
  	xlo = reg_sp & 0xff;
  	xhi = reg_sp >> 8;
  	reg_x = xlo + 256*xhi;
	break;

    case TXA :
  	reg_a = reg_x;		// High byte stuff works right?
  	SetNeg(reg_a, mflag);
  	SetZ(reg_a, mflag);
	break;

    case TXS :
  	reg_sp = reg_x;
	break;

    case TXY :
  	reg_y = reg_x;
  	SetNeg(reg_y, xflag);
  	SetZ(reg_y, xflag);
	break;

    case TYA :
  	reg_a = reg_y;
  	SetNeg(reg_a, mflag);
  	SetZ(reg_a, mflag);
	break;

    case TYX :
  	reg_x = reg_y;
  	SetNeg(reg_x, xflag);
  	SetZ(reg_x, xflag);
	break;

    case WAI :
  	NotImplemented();
	break;

    case WDM :
  	NotImplemented();
	break;

    case XBA :
  	blah = alo;
  	alo = ahi;
  	ahi = blah;
  	reg_a = alo + 256*ahi;
	break;

    case XCE :
  	emul = reg_flags & 0x01;
  	reg_flags = (reg_flags & 0xfe) | reg_e;
  	reg_e = emul;
	break;

  }
}

/*
##################################
#
# Utility routines
#
##################################
*/

void PushStack(char val) {
  ram[reg_sp] = val;
  reg_sp -= 1;
  if ((emul==1) && (reg_sp < 0x0100)) {
    reg_sp += 0x0100;
  }
}

unsigned char PullStack() {
  reg_sp += 1;
  if ((emul==1) && (reg_sp >= 0x0200)) {
    reg_sp = 0x0100;
  }
  return ram[reg_sp];
}

/*
#
# NOTE: On the C64, location 01 determines the memory configuration:
#
#   $A000-$BFFF  ROM/RAM
#   $D000-$DFFF  Hardware registers/CharROM/RAM
#   $E000-$FFFF  ROM/RAM
#
# A read from these locations will fetch ram/rom/regs depending on 01.
# A write to a ROM location always writes to the RAM underneath, whereas
# a write to a hardware register does not.
#
# See IsROM etc. for more info on the different mem configs.
#
*/

unsigned int FetchMemA() {
  return FetchMem(mflag);
}
unsigned int FetchMemX() {
  return FetchMem(xflag);
}

unsigned int FetchMem(unsigned char flag) { // $mflag or $xflag
  long bank, addr;

  if (mode == _IMMED_) {
    return address[0]+256*address[1];
  }
  if (mode == _ACCUM_) {		// e.g. ASL
    return reg_a;
  }
  if (IsReg(effective_address)) {	// Just return 0 for hardware regs
    return 0;
  }

  bank = effective_address >> 16;
  addr = effective_address & 0xffff;
  //&CheckRamBank($bank);
  if (IsROM(effective_address)) {
    if (flag == 0) {
      return rom[addr] + 256*rom[addr+1]; // Potential +1 bug
    } else {
      return rom[addr];
    }
  }

  if ((flag | emul) == 0) {
    return ram[effective_address] +
	    256*ram[effective_address+1]; // Potential +1 bug
  } else {
    return ram[effective_address];
  }
}

void StoreMemA() {
  StoreMem(reg_a, mflag);
}
void StoreMemX() {
  StoreMem(reg_x, xflag);
}
void StoreMemY() {
  StoreMem(reg_y, xflag);
}
void StoreMem(unsigned int val, unsigned char flag) {
  long bank, addr, lo, hi;

  if (flag == 1) {		// mflag or xflag
    val = val & 0xffff;
  } else {
    val = val & 0xff;
  }

  if (IsReg(effective_address)) {
    return;			// Just throw it away
  }
  if (mode == _ACCUM_) {	// e.g. ASL
    reg_a = val;
    return;
  }
  bank = effective_address >> 16;
  addr = effective_address & 0xffff;
  lo = val & 0xff;
  hi = val >> 8;
  //&CheckRamBank($bank);
  ram[effective_address] = (unsigned char)lo;
  if ((flag | emul) == 0) {
    ram[effective_address+1] = (unsigned char)hi; // Potential +1 bug
  }
}

/*
#
# Memory configurations are documented far above, at @a000/@d000/@e000 def.
#
*/

unsigned char IsROM(long addr) {
  unsigned char bits;

  bits = ram[1] & 7;
  if ((0xa000 <= addr) && (addr <= 0xbfff)) {
    return (a000[bits] == _ROM_);		// ROM #defines in globals.h
  } else if ((0xd000 <= addr) && (addr <= 0xdfff)) {
    return (d000[bits] == _ROM_);
  } else if ((0xe000 <= addr) && (addr <= 0xffff)) {
    return (e000[bits] == _ROM_);
  } else {return 0;}
}

unsigned char IsReg(long addr) {
  unsigned char bits;

  bits = ram[1] & 7;
  if ((0xd000 <= addr) && (addr <= 0xdfff)) {
    return (d000[bits] == _IO_);
  } else {return 0;}
}

/*
#
# Set or clear flags.
#
*/
void SetCarry(long arg) {	// .A or memory; always mflag
  long max;
  max = 256;
  if (mflag == 0) {
    max = 65536;
  }
  if (arg >= max) {
    SetFlag(0x01);
  } else {
    ClearFlag(0x01);
  }
}

/*
#
# According to Raeto West (Programming the Commodore 64), the V flag
# in the 6502 is (incorrectly) set as the inverse of the EOR of the
# sign bits.  I have not tested if this is the case with the 816 (and
# am not sufficiently motivated to test the 6502 either!) but assume it is
# and use it as such.
#
*/

void SetV(long argl, long argr) {	// Always .A/mflag (ADC)
  long arg;
  arg = argl ^ argr;
  if ((emul | mflag) == 1) {
    arg = arg & 0x80;
  } else {
    arg = arg & 0x8000;
  }
  if (arg == 0) {
    SetFlag(0x40);
  } else {
    ClearFlag(0x40);
  }
}

void SetNeg(long arg, char flag) {
  long mask;

  mask = 0x80;
  if (flag == 0) {
    mask = 0x8000;
  }
  if ((arg & mask) != 0) {
    SetFlag(0x80);
  } else {
    ClearFlag(0x80);
  }
}

void SetZ(long arg, char flag) {
  long mask;

  mask = 255;
  if (flag == 0) {
    mask = 65535;
  }
  if ((arg & mask) == 0) {
    SetFlag(0x02);
  } else {
    ClearFlag(0x02);
  }
}

void ClearFlag(unsigned char mask) {
  reg_flags = reg_flags & (mask ^ 0xff);
}

void SetFlag(unsigned char mask) {
  reg_flags = reg_flags | mask;
}

void BranchClear(unsigned char mask) {
  if ((reg_flags & mask) == 0) {
    reg_pc = effective_address;
  }
}

void BranchSet(unsigned char mask) {
  if ((reg_flags & mask) != 0) {
    reg_pc = effective_address;
  }
}

void CompareMem(unsigned int reg, unsigned char flag) {
  long arg;

  arg = reg - FetchMem(flag);
  if (arg < 0) {
    ClearFlag(0x01);  // carry
    SetFlag(0x80);    // neg
    ClearFlag(0x02);  // Z
  } else if (arg > 0) {
    SetFlag(0x01);  // carry
    ClearFlag(0x80);    // neg
    ClearFlag(0x02);  // Z
  } else {
    SetFlag(0x01);  // carry
    ClearFlag(0x80);    // neg
    SetFlag(0x02);  // Z
  }
}

long Neg16(long arg) {			// Just retain lower 16 bits if neg
  if (arg < 0) {
    arg = arg & 0xffff;
  }
  return arg;
}

void NotImplemented() {
  char txt[30];
  int i;

  printf("%s (not implemented) at", opcodes[mnem]);

  ToHex16(txt, oppc);
  printf(" PC=%s", txt);
  ToHex16(txt, reg_a);
  printf(" .A=%s", txt);
  ToHex16(txt, reg_x);
  printf(" .X=%s", txt);
  ToHex16(txt, reg_y);
  printf(" .Y=%s", txt);
  ToHex(txt, reg_flags);
  printf(" SR=%s", txt);
  ToHex(txt, reg_e);
  printf(" e=%s", txt);
  ToHex(txt, ram[1]);
  printf(" 01=%s", txt);
  printf("\n");

  ToHex16(txt, reg_sp);
  printf("                         SP=%s>", txt);
  for (i=0; i<8; i++) {
    ToHex(txt, ram[reg_sp+i]);
    printf("%s ", txt);
  }
  printf("\n");

  exit(0);

}
